import DeepProxy from 'proxy-deep';
import { useQueryClient, createQuery, createMutation, createInfiniteQuery, createQueries, } from '@tanstack/svelte-query';
import { onDestroy } from 'svelte';
function getArrayQueryKey(queryKey, input, type) {
    const arrayPath = (typeof queryKey === 'string' ?
        queryKey === '' ? [] : queryKey.split('.')
        : queryKey);
    if (!input && (!type || type === 'any'))
        // for `utils.invalidate()` to match all queries (including vanilla react-query)
        // we don't want nested array if path is empty, i.e. `[]` instead of `[[]]`
        return arrayPath.length ? [arrayPath] : [];
    return [
        arrayPath,
        {
            ...(typeof input !== 'undefined' && { input: input }),
            ...(type && type !== 'any' && { type: type }),
        },
    ];
}
// createContext
const ContextProcedureNames = {
    client: 'client',
    fetch: 'fetch',
    prefetch: 'prefetch',
    fetchInfinite: 'fetchInfinite',
    prefetchInfinite: 'prefetchInfinite',
    invalidate: 'invalidate',
    refetch: 'refetch',
    reset: 'reset',
    cancel: 'cancel',
    setData: 'setData',
    getData: 'getData',
    setInfiniteData: 'setInfiniteData',
    getInfiniteData: 'getInfiniteData',
};
// Procedures
const ProcedureNames = {
    query: 'createQuery',
    serverQuery: 'createServerQuery',
    infiniteQuery: 'createInfiniteQuery',
    serverInfiniteQuery: 'createServerInfiniteQuery',
    mutate: 'createMutation',
    subscribe: 'createSubscription',
    queryKey: 'getQueryKey',
    context: 'createContext',
    queries: 'createQueries',
    serverQueries: 'createServerQueries',
};
// Implementation
function createQueriesProxy(client) {
    return new DeepProxy({}, {
        get() {
            return this.nest(() => { });
        },
        apply(_target, _thisArg, argList) {
            const target = [...this.path].reduce((client, value) => client[value], client);
            const [input, opts] = argList;
            return {
                ...opts,
                queryKey: getArrayQueryKey(this.path, input, 'query'),
                queryFn: () => target.query(input),
            };
        },
    });
}
const contextProcedures = {
    [ContextProcedureNames.fetch]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.fetchQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, 'query'),
                queryFn: () => target.query(input),
            });
        };
    },
    [ContextProcedureNames.prefetch]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.prefetchQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, 'query'),
                queryFn: () => target.query(input),
            });
        };
    },
    [ContextProcedureNames.fetchInfinite]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.fetchInfiniteQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, 'infinite'),
                queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
            });
        };
    },
    [ContextProcedureNames.prefetchInfinite]: ({ path, queryClient, target }) => {
        return (input, opts) => {
            return queryClient.prefetchInfiniteQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, 'infinite'),
                queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
            });
        };
    },
    [ContextProcedureNames.invalidate]: ({ path, queryClient }) => {
        return (input, filters, options) => {
            return queryClient.invalidateQueries({
                ...filters,
                queryKey: getArrayQueryKey(path, input, 'any'),
            }, options);
        };
    },
    [ContextProcedureNames.refetch]: ({ path, queryClient }) => {
        return (input, filters, options) => {
            return queryClient.refetchQueries({
                ...filters,
                queryKey: getArrayQueryKey(path, input, 'any'),
            }, options);
        };
    },
    [ContextProcedureNames.cancel]: ({ path, queryClient }) => {
        return (input, filters, options) => {
            return queryClient.cancelQueries(getArrayQueryKey(path, input, 'any'), filters, options);
        };
    },
    [ContextProcedureNames.reset]: ({ queryClient, path }) => {
        return (input, filters, options) => {
            return queryClient.resetQueries(getArrayQueryKey(path, input, 'any'), filters, options);
        };
    },
    [ContextProcedureNames.setData]: ({ queryClient, path }) => {
        return (input, updater, options) => {
            return queryClient.setQueryData(getArrayQueryKey(path, input, 'query'), updater, options);
        };
    },
    [ContextProcedureNames.setInfiniteData]: ({ queryClient, path }) => {
        return (input, updater, options) => {
            return queryClient.setQueryData(getArrayQueryKey(path, input, 'infinite'), updater, options);
        };
    },
    [ContextProcedureNames.getData]: ({ queryClient, path }) => {
        return (input, filters) => {
            return queryClient.getQueryData(getArrayQueryKey(path, input, 'query'), filters);
        };
    },
    [ContextProcedureNames.getInfiniteData]: ({ queryClient, path }) => {
        return (input, filters) => {
            return queryClient.getQueryData(getArrayQueryKey(path, input, 'infinite'), filters);
        };
    },
};
function createContextProxy(client, queryClient) {
    return new DeepProxy({}, {
        get(_target, key, _receiver) {
            if (key === ContextProcedureNames.client)
                return client;
            if (Object.hasOwn(contextProcedures, key)) {
                const target = [...this.path].reduce((client, value) => client[value], client);
                return contextProcedures[key]({ path: this.path, target, queryClient });
            }
            return this.nest(() => { });
        },
    });
}
const procedures = {
    [ProcedureNames.queryKey]: ({ path }) => {
        return (input, opts) => getArrayQueryKey(path, input, opts);
    },
    [ProcedureNames.query]: ({ path, target }) => {
        const targetFn = target.query;
        return (input, opts) => {
            return createQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, 'query'),
                queryFn: () => targetFn(input),
            });
        };
    },
    [ProcedureNames.serverQuery]: ({ path, target, queryClient }) => {
        const targetFn = target.query;
        return async (input, opts) => {
            const query = {
                queryKey: getArrayQueryKey(path, input, 'query'),
                queryFn: () => targetFn(input),
            };
            const cache = queryClient.getQueryCache().find(query.queryKey);
            if (opts?.ssr !== false && !cache) {
                await queryClient.prefetchQuery(query);
            }
            return () => createQuery({
                ...opts,
                ...query,
                ...(!cache ?
                    { refetchOnMount: opts?.refetchOnMount ?? false } : {}),
            });
        };
    },
    [ProcedureNames.infiniteQuery]: ({ path, target }) => {
        return (input, opts) => {
            return createInfiniteQuery({
                ...opts,
                queryKey: getArrayQueryKey(path, input, 'infinite'),
                queryFn: ({ pageParam }) => target.query({ ...input, cursor: pageParam }),
            });
        };
    },
    [ProcedureNames.serverInfiniteQuery]: ({ path, target, queryClient }) => {
        const targetFn = target.query;
        return async (input, opts) => {
            const query = {
                queryKey: getArrayQueryKey(path, input, 'infinite'),
                queryFn: ({ pageParam }) => targetFn({ ...input, cursor: pageParam }),
            };
            const cache = queryClient.getQueryCache().find(query.queryKey);
            if (opts?.ssr !== false && !cache) {
                await queryClient.prefetchInfiniteQuery(query);
            }
            return () => createInfiniteQuery({
                ...opts,
                ...query,
                ...(!cache ?
                    { refetchOnMount: opts?.refetchOnMount ?? false } : {}),
            });
        };
    },
    [ProcedureNames.mutate]: ({ path, target }) => {
        return (opts) => {
            return createMutation({
                ...opts,
                mutationKey: path,
                mutationFn: (data) => target.mutate(data),
            });
        };
    },
    [ProcedureNames.subscribe]: ({ target }) => {
        return (input, opts) => {
            const enabled = opts?.enabled ?? true;
            if (!enabled)
                return;
            let isStopped = false;
            const subscription = target.subscribe(input, {
                onStarted: () => {
                    if (!isStopped)
                        opts?.onStarted?.();
                },
                onData: (data) => {
                    if (!isStopped)
                        opts?.onData?.(data);
                },
                onError: (err) => {
                    if (!isStopped)
                        opts?.onError?.(err);
                },
            });
            return onDestroy(() => {
                isStopped = true;
                subscription.unsubscribe();
            });
        };
    },
    [ProcedureNames.queries]: ({ path, queriesProxy }) => {
        if (path.length !== 0)
            return;
        return (input) => {
            const proxy = queriesProxy();
            return createQueries(input(proxy));
        };
    },
    [ProcedureNames.serverQueries]: ({ path, queriesProxy, queryClient }) => {
        if (path.length !== 0)
            return;
        const proxy = queriesProxy();
        return async (input) => {
            const queryKeys = await Promise.all(input(proxy).map(async (query) => {
                const cache = queryClient.getQueryCache().find(query.queryKey);
                if (query.ssr !== false && !cache) {
                    await queryClient.prefetchQuery(query);
                }
                return {
                    ...query,
                    ...(!cache ?
                        { refetchOnMount: query.refetchOnMount ?? false } : {}),
                };
            }));
            return () => createQueries(queryKeys);
        };
    },
    [ProcedureNames.context]: ({ path, contextProxy }) => {
        if (path.length !== 0)
            return;
        return contextProxy;
    },
};
export function svelteQueryWrapper({ client, queryClient, }) {
    const qc = queryClient ?? useQueryClient();
    return new DeepProxy({}, {
        get(_, key) {
            if (Object.hasOwn(procedures, key)) {
                const target = [...this.path].reduce((client, value) => client[value], client);
                return procedures[key]({
                    path: this.path,
                    target,
                    queryClient: qc,
                    queriesProxy: () => createQueriesProxy(client),
                    contextProxy: () => createContextProxy(client, qc),
                });
            }
            return this.nest(() => { });
        },
    });
}
